@@ -24,11 +24,14 @@
 #include "Logging.hpp"
 #include "TemporaryFile.hpp"
 #include "Win32Util.hpp"
-#include "fmtmacros.hpp"
 
+#include <Config.hpp>
 #include <Finalizer.hpp>
 #include <core/exceptions.hpp>
 #include <core/wincompat.hpp>
+#include <fmtmacros.hpp>
+#include <util/TimePoint.hpp>
+#include <util/file.hpp>
 #include <util/path.hpp>
 #include <util/string.hpp>
 
@@ -46,32 +49,10 @@ extern "C" {
 
 #include <fcntl.h>
 
-#include <algorithm>
-#include <climits>
-#include <codecvt>
-#include <fstream>
-#include <locale>
-
-#ifndef HAVE_DIRENT_H
-#  include <filesystem>
-#endif
-
 #ifdef HAVE_PWD_H
 #  include <pwd.h>
 #endif
 
-#ifdef HAVE_SYS_TIME_H
-#  include <sys/time.h>
-#endif
-
-#ifdef HAVE_LINUX_FS_H
-#  include <linux/magic.h>
-#  include <sys/statfs.h>
-#elif defined(HAVE_STRUCT_STATFS_F_FSTYPENAME)
-#  include <sys/mount.h>
-#  include <sys/param.h>
-#endif
-
 #ifdef __linux__
 #  ifdef HAVE_SYS_IOCTL_H
 #    include <sys/ioctl.h>
@@ -94,9 +75,6 @@ extern "C" {
 #  endif
 #endif
 
-using nonstd::nullopt;
-using nonstd::optional;
-using nonstd::string_view;
 using IncludeDelimiter = util::Tokenizer::IncludeDelimiter;
 
 namespace {
@@ -115,8 +93,8 @@ mode_t g_umask = [] {
 // The primary reason for not using std::regex is that it's not available for
 // GCC 4.8. It's also a bit bloated. The reason for not using POSIX regex
 // functionality is that it's are not available in MinGW.
-string_view
-find_first_ansi_csi_seq(string_view string)
+std::string_view
+find_first_ansi_csi_seq(std::string_view string)
 {
   size_t pos = 0;
   while (pos < string.length() && string[pos] != 0x1b) {
@@ -159,7 +137,7 @@ path_max(const std::string& path)
 
 template<typename T>
 std::vector<T>
-split_into(string_view string,
+split_into(std::string_view string,
            const char* separators,
            util::Tokenizer::Mode mode,
            IncludeDelimiter include_delimiter)
@@ -174,7 +152,7 @@ split_into(string_view string,
 }
 
 std::string
-rewrite_stderr_to_absolute_paths(string_view text)
+rewrite_stderr_to_absolute_paths(std::string_view text)
 {
   static const std::string in_file_included_from = "In file included from ";
 
@@ -200,7 +178,7 @@ rewrite_stderr_to_absolute_paths(string_view text)
       line = line.substr(csi_seq.length());
     }
     size_t path_end = line.find(':');
-    if (path_end == string_view::npos) {
+    if (path_end == std::string_view::npos) {
       result.append(line.data(), line.length());
     } else {
       std::string path(line.substr(0, path_end));
@@ -216,22 +194,12 @@ rewrite_stderr_to_absolute_paths(string_view text)
   return result;
 }
 
-#ifdef _WIN32
-bool
-has_utf16_le_bom(string_view text)
-{
-  return text.size() > 1
-         && ((static_cast<uint8_t>(text[0]) == 0xff
-              && static_cast<uint8_t>(text[1]) == 0xfe));
-}
-#endif
-
 } // namespace
 
 namespace Util {
 
-string_view
-base_name(string_view path)
+std::string_view
+base_name(std::string_view path)
 {
 #ifdef _WIN32
   const char delim[] = "/\\";
@@ -243,9 +211,9 @@ base_name(string_view path)
 }
 
 std::string
-change_extension(string_view path, string_view new_ext)
+change_extension(std::string_view path, std::string_view new_ext)
 {
-  string_view without_ext = Util::remove_extension(path);
+  std::string_view without_ext = Util::remove_extension(path);
   return std::string(without_ext).append(new_ext.data(), new_ext.length());
 }
 
@@ -256,7 +224,7 @@ clone_file(const std::string& src, const std::string& dest, bool via_tmp_file)
 #  if defined(__linux__)
   Fd src_fd(open(src.c_str(), O_RDONLY));
   if (!src_fd) {
-    throw core::Error("{}: {}", src, strerror(errno));
+    throw core::Error(FMT("{}: {}", src, strerror(errno)));
   }
 
   Fd dest_fd;
@@ -269,7 +237,7 @@ clone_file(const std::string& src, const std::string& dest, bool via_tmp_file)
     dest_fd =
       Fd(open(dest.c_str(), O_WRONLY | O_CREAT | O_TRUNC | O_BINARY, 0666));
     if (!dest_fd) {
-      throw core::Error("{}: {}", src, strerror(errno));
+      throw core::Error(FMT("{}: {}", src, strerror(errno)));
     }
   }
 
@@ -298,12 +266,12 @@ clone_file(const std::string& src, const std::string& dest, bool via_tmp_file)
 #endif // FILE_CLONING_SUPPORTED
 
 void
-clone_hard_link_or_copy_file(const Context& ctx,
+clone_hard_link_or_copy_file(const Config& config,
                              const std::string& source,
                              const std::string& dest,
                              bool via_tmp_file)
 {
-  if (ctx.config.file_clone()) {
+  if (config.file_clone()) {
 #ifdef FILE_CLONING_SUPPORTED
     LOG("Cloning {} to {}", source, dest);
     try {
@@ -316,18 +284,18 @@ clone_hard_link_or_copy_file(const Context& ctx,
     LOG("Not cloning {} to {} since it's unsupported", source, dest);
 #endif
   }
-  if (ctx.config.hard_link()) {
+  if (config.hard_link()) {
     LOG("Hard linking {} to {}", source, dest);
     try {
       Util::hard_link(source, dest);
 #ifndef _WIN32
       if (chmod(dest.c_str(), 0444 & ~Util::get_umask()) != 0) {
-        LOG("Failed to chmod: {}", strerror(errno));
+        LOG("Failed to chmod {}: {}", dest.c_str(), strerror(errno));
       }
 #endif
       return;
     } catch (const core::Error& e) {
-      LOG_RAW(e.what());
+      LOG("Failed to hard link {} to {}: {}", source, dest, e.what());
       // Fall back to copying.
     }
   }
@@ -337,7 +305,7 @@ clone_hard_link_or_copy_file(const Context& ctx,
 }
 
 size_t
-common_dir_prefix_length(string_view dir, string_view path)
+common_dir_prefix_length(std::string_view dir, std::string_view path)
 {
   if (dir.empty() || path.empty() || dir == "/" || path == "/") {
     return 0;
@@ -369,8 +337,9 @@ common_dir_prefix_length(string_view dir, string_view path)
 void
 copy_fd(int fd_in, int fd_out)
 {
-  read_fd(fd_in,
-          [=](const void* data, size_t size) { write_fd(fd_out, data, size); });
+  util::read_fd(fd_in, [=](const void* data, size_t size) {
+    util::write_fd(fd_out, data, size);
+  });
 }
 
 void
@@ -378,9 +347,12 @@ copy_file(const std::string& src, const std::string& dest, bool via_tmp_file)
 {
   Fd src_fd(open(src.c_str(), O_RDONLY | O_BINARY));
   if (!src_fd) {
-    throw core::Error("{}: {}", src, strerror(errno));
+    throw core::Error(
+      FMT("Failed to open {} for reading: {}", src, strerror(errno)));
   }
 
+  unlink(dest.c_str());
+
   Fd dest_fd;
   std::string tmp_file;
   if (via_tmp_file) {
@@ -391,7 +363,8 @@ copy_file(const std::string& src, const std::string& dest, bool via_tmp_file)
     dest_fd =
       Fd(open(dest.c_str(), O_WRONLY | O_CREAT | O_TRUNC | O_BINARY, 0666));
     if (!dest_fd) {
-      throw core::Error("{}: {}", dest, strerror(errno));
+      throw core::Error(
+        FMT("Failed to open {} for writing: {}", dest, strerror(errno)));
     }
   }
 
@@ -405,7 +378,7 @@ copy_file(const std::string& src, const std::string& dest, bool via_tmp_file)
 }
 
 bool
-create_dir(string_view dir)
+create_dir(std::string_view dir)
 {
   std::string dir_str(dir);
   auto st = Stat::stat(dir_str);
@@ -432,8 +405,8 @@ create_dir(string_view dir)
   }
 }
 
-string_view
-dir_name(string_view path)
+std::string_view
+dir_name(std::string_view path)
 {
 #ifdef _WIN32
   const char delim[] = "/\\";
@@ -477,7 +450,7 @@ expand_environment_variables(const std::string& str)
         ++right;
       }
       if (curly && *right != '}') {
-        throw core::Error("syntax error: missing '}}' after \"{}\"", left);
+        throw core::Error(FMT("syntax error: missing '}}' after \"{}\"", left));
       }
       if (right == left) {
         // Special case: don't consider a single $ the left of a variable.
@@ -487,7 +460,7 @@ expand_environment_variables(const std::string& str)
         std::string name(left, right - left);
         const char* value = getenv(name.c_str());
         if (!value) {
-          throw core::Error("environment variable \"{}\" not set", name);
+          throw core::Error(FMT("environment variable \"{}\" not set", name));
         }
         result += value;
         if (!curly) {
@@ -526,7 +499,7 @@ fallocate(int fd, long new_size)
   }
   int err = 0;
   try {
-    write_fd(fd, buf, bytes_to_write);
+    util::write_fd(fd, buf, bytes_to_write);
   } catch (core::Error&) {
     err = errno;
   }
@@ -599,11 +572,11 @@ format_parsable_size_with_suffix(uint64_t size)
 }
 
 void
-ensure_dir_exists(nonstd::string_view dir)
+ensure_dir_exists(std::string_view dir)
 {
   if (!create_dir(dir)) {
     throw core::Fatal(
-      "Failed to create directory {}: {}", dir, strerror(errno));
+      FMT("Failed to create directory {}: {}", dir, strerror(errno)));
   }
 }
 
@@ -643,8 +616,8 @@ get_apparent_cwd(const std::string& actual_cwd)
 #endif
 }
 
-string_view
-get_extension(string_view path)
+std::string_view
+get_extension(std::string_view path)
 {
 #ifndef _WIN32
   const char stop_at_chars[] = "./";
@@ -652,7 +625,7 @@ get_extension(string_view path)
   const char stop_at_chars[] = "./\\";
 #endif
   size_t pos = path.find_last_of(stop_at_chars);
-  if (pos == string_view::npos || path.at(pos) == '/') {
+  if (pos == std::string_view::npos || path.at(pos) == '/') {
     return {};
 #ifdef _WIN32
   } else if (path.at(pos) == '\\') {
@@ -666,26 +639,28 @@ get_extension(string_view path)
 std::string
 get_home_directory()
 {
-  const char* p = getenv("HOME");
-  if (p) {
-    return p;
-  }
 #ifdef _WIN32
-  p = getenv("APPDATA");
-  if (p) {
+  if (const char* p = getenv("USERPROFILE")) {
     return p;
   }
-#endif
-#ifdef HAVE_GETPWUID
+  throw core::Fatal(
+    "The USERPROFILE environment variable must be set to your user profile "
+    "folder");
+#else
+  if (const char* p = getenv("HOME")) {
+    return p;
+  }
+#  ifdef HAVE_GETPWUID
   {
     struct passwd* pwd = getpwuid(getuid());
     if (pwd) {
       return pwd->pw_dir;
     }
   }
-#endif
+#  endif
   throw core::Fatal(
     "Could not determine home directory from $HOME or getpwuid(3)");
+#endif
 }
 
 const char*
@@ -705,7 +680,7 @@ get_hostname()
 }
 
 std::string
-get_relative_path(string_view dir, string_view path)
+get_relative_path(std::string_view dir, std::string_view path)
 {
   ASSERT(util::is_absolute_path(dir));
   ASSERT(util::is_absolute_path(path));
@@ -764,22 +739,17 @@ hard_link(const std::string& oldpath, const std::string& newpath)
 
 #ifndef _WIN32
   if (link(oldpath.c_str(), newpath.c_str()) != 0) {
-    throw core::Error(
-      "failed to link {} to {}: {}", oldpath, newpath, strerror(errno));
+    throw core::Error(strerror(errno));
   }
 #else
   if (!CreateHardLink(newpath.c_str(), oldpath.c_str(), nullptr)) {
-    DWORD error = GetLastError();
-    throw core::Error("failed to link {} to {}: {}",
-                      oldpath,
-                      newpath,
-                      Win32Util::error_message(error));
+    throw core::Error(Win32Util::error_message(GetLastError()));
   }
 #endif
 }
 
-nonstd::optional<size_t>
-is_absolute_path_with_prefix(nonstd::string_view path)
+std::optional<size_t>
+is_absolute_path_with_prefix(std::string_view path)
 {
 #ifdef _WIN32
   const char delim[] = "/\\";
@@ -800,49 +770,36 @@ is_absolute_path_with_prefix(nonstd::string_view path)
     // NOLINTNEXTLINE(readability-simplify-boolean-expr)
     return split_pos;
   }
-  return nonstd::nullopt;
+  return std::nullopt;
 }
 
-#if defined(HAVE_LINUX_FS_H) || defined(HAVE_STRUCT_STATFS_F_FSTYPENAME)
-int
-is_nfs_fd(int fd, bool* is_nfs)
-{
-  struct statfs buf;
-  if (fstatfs(fd, &buf) != 0) {
-    return errno;
-  }
-#  ifdef HAVE_LINUX_FS_H
-  *is_nfs = buf.f_type == NFS_SUPER_MAGIC;
-#  else // Mac OS X and some other BSD flavors
-  *is_nfs = strcmp(buf.f_fstypename, "nfs") == 0;
-#  endif
-  return 0;
-}
-#else
-int
-is_nfs_fd(int /*fd*/, bool* /*is_nfs*/)
-{
-  return -1;
-}
-#endif
-
 bool
-is_precompiled_header(string_view path)
+is_ccache_executable(const std::string_view path)
 {
-  string_view ext = get_extension(path);
+  std::string name(Util::base_name(path));
+#ifdef _WIN32
+  name = Util::to_lowercase(name);
+#endif
+  return util::starts_with(name, "ccache");
+}
+
+bool
+is_precompiled_header(std::string_view path)
+{
+  std::string_view ext = get_extension(path);
   return ext == ".gch" || ext == ".pch" || ext == ".pth"
          || get_extension(dir_name(path)) == ".gch";
 }
 
-optional<tm>
-localtime(optional<time_t> time)
+std::optional<tm>
+localtime(std::optional<util::TimePoint> time)
 {
-  time_t timestamp = time ? *time : ::time(nullptr);
+  time_t timestamp = time ? time->sec() : util::TimePoint::now().sec();
   tm result;
   if (localtime_r(&timestamp, &result)) {
     return result;
   } else {
-    return nullopt;
+    return std::nullopt;
   }
 }
 
@@ -850,7 +807,7 @@ std::string
 make_relative_path(const std::string& base_dir,
                    const std::string& actual_cwd,
                    const std::string& apparent_cwd,
-                   nonstd::string_view path)
+                   std::string_view path)
 {
   if (base_dir.empty() || !util::path_starts_with(path, base_dir)) {
     return std::string(path);
@@ -916,32 +873,33 @@ make_relative_path(const std::string& base_dir,
 }
 
 std::string
-make_relative_path(const Context& ctx, string_view path)
+make_relative_path(const Context& ctx, std::string_view path)
 {
   return make_relative_path(
     ctx.config.base_dir(), ctx.actual_cwd, ctx.apparent_cwd, path);
 }
 
 bool
-matches_dir_prefix_or_file(string_view dir_prefix_or_file, string_view path)
+matches_dir_prefix_or_file(std::string_view dir_prefix_or_file,
+                           std::string_view path)
 {
   return !dir_prefix_or_file.empty() && !path.empty()
          && dir_prefix_or_file.length() <= path.length()
-         && path.starts_with(dir_prefix_or_file)
+         && util::starts_with(path, dir_prefix_or_file)
          && (dir_prefix_or_file.length() == path.length()
              || is_dir_separator(path[dir_prefix_or_file.length()])
              || is_dir_separator(dir_prefix_or_file.back()));
 }
 
 std::string
-normalize_abstract_absolute_path(string_view path)
+normalize_abstract_absolute_path(std::string_view path)
 {
   if (!util::is_absolute_path(path)) {
     return std::string(path);
   }
 
 #ifdef _WIN32
-  if (path.find("\\") != string_view::npos) {
+  if (path.find("\\") != std::string_view::npos) {
     std::string new_path(path);
     std::replace(new_path.begin(), new_path.end(), '\\', '/');
     return normalize_abstract_absolute_path(new_path);
@@ -952,7 +910,7 @@ normalize_abstract_absolute_path(string_view path)
 #endif
 
   std::string result = "/";
-  const size_t npos = string_view::npos;
+  const size_t npos = std::string_view::npos;
   size_t left = 1;
 
   while (true) {
@@ -960,7 +918,8 @@ normalize_abstract_absolute_path(string_view path)
       break;
     }
     const auto right = path.find('/', left);
-    string_view part = path.substr(left, right == npos ? npos : right - left);
+    std::string_view part =
+      path.substr(left, right == npos ? npos : right - left);
     if (part == "..") {
       if (result.length() > 1) {
         // "/x/../part" -> "/part"
@@ -1002,7 +961,7 @@ normalize_concrete_absolute_path(const std::string& path)
 }
 
 uint64_t
-parse_duration(const std::string& duration)
+parse_duration(std::string_view duration)
 {
   uint64_t factor = 0;
   char last_ch = duration.empty() ? '\0' : duration[duration.length() - 1];
@@ -1015,8 +974,8 @@ parse_duration(const std::string& duration)
     factor = 1;
     break;
   default:
-    throw core::Error(
-      "invalid suffix (supported: d (day) and s (second)): \"{}\"", duration);
+    throw core::Error(FMT(
+      "invalid suffix (supported: d (day) and s (second)): \"{}\"", duration));
   }
 
   const auto value =
@@ -1036,7 +995,7 @@ parse_size(const std::string& value)
   char* p;
   double result = strtod(value.c_str(), &p);
   if (errno != 0 || result < 0 || p == value.c_str() || value.empty()) {
-    throw core::Error("invalid size: \"{}\"", value);
+    throw core::Error(FMT("invalid size: \"{}\"", value));
   }
 
   while (isspace(*p)) {
@@ -1048,19 +1007,19 @@ parse_size(const std::string& value)
     switch (*p) {
     case 'T':
       result *= multiplier;
-    // Fallthrough.
+      [[fallthrough]];
     case 'G':
       result *= multiplier;
-    // Fallthrough.
+      [[fallthrough]];
     case 'M':
       result *= multiplier;
-    // Fallthrough.
+      [[fallthrough]];
     case 'K':
     case 'k':
       result *= multiplier;
       break;
     default:
-      throw core::Error("invalid size: \"{}\"", value);
+      throw core::Error(FMT("invalid size: \"{}\"", value));
     }
   } else {
     // Default suffix: G.
@@ -1069,72 +1028,6 @@ parse_size(const std::string& value)
   return static_cast<uint64_t>(result);
 }
 
-bool
-read_fd(int fd, DataReceiver data_receiver)
-{
-  int64_t n;
-  char buffer[CCACHE_READ_BUFFER_SIZE];
-  while ((n = read(fd, buffer, sizeof(buffer))) != 0) {
-    if (n == -1 && errno != EINTR) {
-      break;
-    }
-    if (n > 0) {
-      data_receiver(buffer, n);
-    }
-  }
-  return n >= 0;
-}
-
-std::string
-read_file(const std::string& path, size_t size_hint)
-{
-  if (size_hint == 0) {
-    auto stat = Stat::stat(path);
-    if (!stat) {
-      throw core::Error(strerror(errno));
-    }
-    size_hint = stat.size();
-  }
-
-  // +1 to be able to detect EOF in the first read call
-  size_hint = (size_hint < 1024) ? 1024 : size_hint + 1;
-
-  Fd fd(open(path.c_str(), O_RDONLY | O_BINARY));
-  if (!fd) {
-    throw core::Error(strerror(errno));
-  }
-
-  int64_t ret = 0;
-  size_t pos = 0;
-  std::string result;
-  result.resize(size_hint);
-
-  while (true) {
-    if (pos == result.size()) {
-      result.resize(2 * result.size());
-    }
-    const size_t max_read = result.size() - pos;
-    ret = read(*fd, &result[pos], max_read);
-    if (ret == 0 || (ret == -1 && errno != EINTR)) {
-      break;
-    }
-    if (ret > 0) {
-      pos += ret;
-      if (static_cast<size_t>(ret) < max_read) {
-        break;
-      }
-    }
-  }
-
-  if (ret == -1) {
-    LOG("Failed reading {}", path);
-    throw core::Error(strerror(errno));
-  }
-
-  result.resize(pos);
-  return result;
-}
-
 #ifndef _WIN32
 std::string
 read_link(const std::string& path)
@@ -1191,29 +1084,8 @@ real_path(const std::string& path, bool return_empty_on_error)
   return resolved ? resolved : (return_empty_on_error ? "" : path);
 }
 
-std::string
-read_text_file(const std::string& path, size_t size_hint)
-{
-  std::string result = read_file(path, size_hint);
-#ifdef _WIN32
-  // Convert to UTF-8 if the content starts with a UTF-16 little-endian BOM.
-  //
-  // Note that this code assumes a little-endian machine, which is why it's
-  // #ifdef-ed to only run on Windows (which is always little-endian) where it's
-  // actually needed.
-  if (has_utf16_le_bom(result)) {
-    result.erase(0, 2); // Remove BOM.
-    std::u16string result_as_u16((result.size() / 2) + 1, '\0');
-    result_as_u16 = reinterpret_cast<const char16_t*>(result.c_str());
-    std::wstring_convert<std::codecvt_utf8_utf16<char16_t>, char16_t> converter;
-    result = converter.to_bytes(result_as_u16);
-  }
-#endif
-  return result;
-}
-
-string_view
-remove_extension(string_view path)
+std::string_view
+remove_extension(std::string_view path)
 {
   return path.substr(0, path.length() - get_extension(path).length());
 }
@@ -1224,7 +1096,7 @@ rename(const std::string& oldpath, const std::string& newpath)
 #ifndef _WIN32
   if (::rename(oldpath.c_str(), newpath.c_str()) != 0) {
     throw core::Error(
-      "failed to rename {} to {}: {}", oldpath, newpath, strerror(errno));
+      FMT("failed to rename {} to {}: {}", oldpath, newpath, strerror(errno)));
   }
 #else
   // Windows' rename() won't overwrite an existing file, so need to use
@@ -1232,31 +1104,18 @@ rename(const std::string& oldpath, const std::string& newpath)
   if (!MoveFileExA(
         oldpath.c_str(), newpath.c_str(), MOVEFILE_REPLACE_EXISTING)) {
     DWORD error = GetLastError();
-    throw core::Error("failed to rename {} to {}: {}",
-                      oldpath,
-                      newpath,
-                      Win32Util::error_message(error));
+    throw core::Error(FMT("failed to rename {} to {}: {}",
+                          oldpath,
+                          newpath,
+                          Win32Util::error_message(error)));
   }
 #endif
 }
 
-bool
-same_program_name(nonstd::string_view program_name,
-                  nonstd::string_view canonical_program_name)
-{
-#ifdef _WIN32
-  std::string lowercase_program_name = Util::to_lowercase(program_name);
-  return lowercase_program_name == canonical_program_name
-         || lowercase_program_name == FMT("{}.exe", canonical_program_name);
-#else
-  return program_name == canonical_program_name;
-#endif
-}
-
 void
-send_to_fd(const Context& ctx, const std::string& text, int fd)
+send_to_fd(const Context& ctx, std::string_view text, int fd)
 {
-  const std::string* text_to_send = &text;
+  std::string_view text_to_send = text;
   std::string modified_text;
 
 #ifdef _WIN32
@@ -1270,21 +1129,21 @@ send_to_fd(const Context& ctx, const std::string& text, int fd)
   if (ctx.args_info.strip_diagnostics_colors) {
     try {
       modified_text = strip_ansi_csi_seqs(text);
-      text_to_send = &modified_text;
+      text_to_send = modified_text;
     } catch (const core::Error&) {
-      // Fall through
+      // Ignore.
     }
   }
 
   if (ctx.config.absolute_paths_in_stderr()) {
-    modified_text = rewrite_stderr_to_absolute_paths(*text_to_send);
-    text_to_send = &modified_text;
+    modified_text = rewrite_stderr_to_absolute_paths(text_to_send);
+    text_to_send = modified_text;
   }
 
-  try {
-    write_fd(fd, text_to_send->data(), text_to_send->length());
-  } catch (core::Error& e) {
-    throw core::Error("Failed to write to {}: {}", fd, e.what());
+  const auto result =
+    util::write_fd(fd, text_to_send.data(), text_to_send.length());
+  if (!result) {
+    throw core::Error(FMT("Failed to write to {}: {}", fd, result.error()));
   }
 }
 
@@ -1320,17 +1179,18 @@ setenv(const std::string& name, const std::string& value)
 #endif
 }
 
-std::vector<string_view>
-split_into_views(string_view string,
+std::vector<std::string_view>
+split_into_views(std::string_view string,
                  const char* separators,
                  util::Tokenizer::Mode mode,
                  IncludeDelimiter include_delimiter)
 {
-  return split_into<string_view>(string, separators, mode, include_delimiter);
+  return split_into<std::string_view>(
+    string, separators, mode, include_delimiter);
 }
 
 std::vector<std::string>
-split_into_strings(string_view string,
+split_into_strings(std::string_view string,
                    const char* separators,
                    util::Tokenizer::Mode mode,
                    IncludeDelimiter include_delimiter)
@@ -1339,7 +1199,7 @@ split_into_strings(string_view string,
 }
 
 std::string
-strip_ansi_csi_seqs(string_view string)
+strip_ansi_csi_seqs(std::string_view string)
 {
   size_t pos = 0;
   std::string result;
@@ -1361,7 +1221,7 @@ strip_ansi_csi_seqs(string_view string)
 }
 
 std::string
-to_lowercase(string_view string)
+to_lowercase(std::string_view string)
 {
   std::string result;
   result.resize(string.length());
@@ -1396,9 +1256,9 @@ traverse(const std::string& path, const TraverseVisitor& visitor)
           if (stat.error_number() == ENOENT || stat.error_number() == ESTALE) {
             continue;
           }
-          throw core::Error("failed to lstat {}: {}",
-                            entry_path,
-                            strerror(stat.error_number()));
+          throw core::Error(FMT("failed to lstat {}: {}",
+                                entry_path,
+                                strerror(stat.error_number())));
         }
         is_dir = stat.is_directory();
       }
@@ -1413,7 +1273,8 @@ traverse(const std::string& path, const TraverseVisitor& visitor)
   } else if (errno == ENOTDIR) {
     visitor(path, false);
   } else {
-    throw core::Error("failed to open directory {}: {}", path, strerror(errno));
+    throw core::Error(
+      FMT("failed to open directory {}: {}", path, strerror(errno)));
   }
 }
 
@@ -1436,7 +1297,8 @@ traverse(const std::string& path, const TraverseVisitor& visitor)
   } else if (std::filesystem::exists(path)) {
     visitor(path, false);
   } else {
-    throw core::Error("failed to open directory {}: {}", path, strerror(errno));
+    throw core::Error(
+      FMT("failed to open directory {}: {}", path, strerror(errno)));
   }
 }
 
@@ -1450,7 +1312,8 @@ unlink_safe(const std::string& path, UnlinkLog unlink_log)
   // If path is on an NFS share, unlink isn't atomic, so we rename to a temp
   // file. We don't care if the temp file is trashed, so it's always safe to
   // unlink it first.
-  std::string tmp_name = path + ".ccache.rm.tmp";
+  const std::string tmp_name =
+    FMT("{}.ccache{}unlink", path, TemporaryFile::tmp_file_infix);
 
   bool success = true;
   try {
@@ -1501,6 +1364,8 @@ unsetenv(const std::string& name)
 {
 #ifdef HAVE_UNSETENV
   ::unsetenv(name.c_str());
+#elif defined(_WIN32)
+  SetEnvironmentVariable(name.c_str(), NULL);
 #else
   putenv(strdup(name.c_str())); // Leak to environment.
 #endif
@@ -1515,46 +1380,12 @@ wipe_path(const std::string& path)
   traverse(path, [](const std::string& p, bool is_dir) {
     if (is_dir) {
       if (rmdir(p.c_str()) != 0 && errno != ENOENT && errno != ESTALE) {
-        throw core::Error("failed to rmdir {}: {}", p, strerror(errno));
+        throw core::Error(FMT("failed to rmdir {}: {}", p, strerror(errno)));
       }
     } else if (unlink(p.c_str()) != 0 && errno != ENOENT && errno != ESTALE) {
-      throw core::Error("failed to unlink {}: {}", p, strerror(errno));
+      throw core::Error(FMT("failed to unlink {}: {}", p, strerror(errno)));
     }
   });
 }
 
-void
-write_fd(int fd, const void* data, size_t size)
-{
-  int64_t written = 0;
-  do {
-    const auto count =
-      write(fd, static_cast<const uint8_t*>(data) + written, size - written);
-    if (count == -1) {
-      if (errno != EAGAIN && errno != EINTR) {
-        throw core::Error(strerror(errno));
-      }
-    } else {
-      written += count;
-    }
-  } while (static_cast<size_t>(written) < size);
-}
-
-void
-write_file(const std::string& path,
-           const std::string& data,
-           std::ios_base::openmode open_mode)
-{
-  if (path.empty()) {
-    throw core::Error("No such file or directory");
-  }
-
-  open_mode |= std::ios::out;
-  std::ofstream file(path, open_mode);
-  if (!file) {
-    throw core::Error(strerror(errno));
-  }
-  file << data;
-}
-
 } // namespace Util
